
//////////////////////////////////
//                              //
//  FindOptimalSegmentations.c  //
//                              //
//////////////////////////////////

#include "FindOptimalSegmentations.h"

#include <stdlib.h>
#include <math.h>

int FindOptimalSegmentations(const double * x, const double * y, const unsigned nMax, const unsigned kMax, unsigned * segBoundaries, double * segSumErrorSquared)
{
    // Note: in this function, the individual segment numberings are 1-based; they range from [1 .. nMax].

    // Verify parameter values. We need 1 <= kMax <= nMax.

    if (!((1 <= kMax) && (kMax <= nMax)))
    {
        return FindOptimalSegmentation_ERROR_ParameterValue; // report parameter error
    }

    // Allocate (kMax, nMax) matrix for recording the estimated-error-squared of optimal (sub)solutions

    double * sumErrorSquared = (double *)malloc(kMax * nMax * sizeof(double));

    if (sumErrorSquared == NULL)
    {
        return FindOptimalSegmentation_ERROR_OutOfMemory; // report out of memory
    }

    // Allocate (kMax, nMax) matrix for tracing optimal (sub)solutions

    unsigned * trace = (unsigned *)malloc(kMax * nMax * sizeof(unsigned));

    if (trace == NULL)
    {
        free(sumErrorSquared); // release previously allocated sumErrorSquared
        return FindOptimalSegmentation_ERROR_OutOfMemory; // report out of memory
    }

    // Initialize arrays

    for (unsigned k = 1; k <= kMax; ++k)
    {
        for (unsigned n = 1; n <= nMax; ++n)
        {
            sumErrorSquared[(k - 1) * nMax + (n - 1)] = INFINITY;
            trace          [(k - 1) * nMax + (n - 1)] = 0;
        }
    }

    // Calculate optimal segmentations using dynamic programming.

    for (unsigned a = 1; a <= nMax; ++a)
    {
        // Invariant: for all segments [1 .. a - 1] and all segment-counts 'k',
        // the 'errorSquared' and 'trace' tables reflect the optimal solution.
        //
        // The goal of this a-loop iteration is to make this true for
        // all segments [1 .. a].

        double xx = 0.0;
        double xy = 0.0;
        double yy = 0.0;

        for (unsigned b = a; b <= nMax; ++b)
        {
            // We will calculate the value of a (sub)solution that has [a .. b] as its last (rightmost) segment.

            xx += x[b - 1] * x[b - 1]; // now equal to SUM(x[i] * x[i], i = a..b)
            xy += x[b - 1] * y[b - 1]; // now equal to SUM(x[i] * y[i], i = a..b)
            yy += y[b - 1] * y[b - 1]; // now equal to SUM(y[i] * y[i], i = a..b)

            // Calculate the error-squared contribution of the segment [a .. b]
            //
            // Note that, if a segment has only zero x[i] values, we can still give
            // an error estimate.

            const double sumErrorSquaredCurrentStep = (xx == 0.0) ? yy : (yy - xy * xy / xx);

            // We will now consider updating errorSquared[k, b] with a solution that has
            // 'k' segments, the current segment [a .. b] as its last segment.
            //
            // Either the segment [a ..b] is the first segment, or it is adjoined
            // to the segment described by errorSquared[k - 1, a - 1].

            const unsigned kHi = (a <= kMax) ? a : kMax;

            //#pragma omp parallel for
            for (unsigned k = 1; k <= kHi ; ++k)
            {
                // If a == 1, this "last" step is also the "first" step; thus, errorSquaredPreviousSteps == 0.
                // The cost of the previous steps is made Infinite for (k == 1, a > 1). This prevents these solutions
                // being considered; solutions for k = 1 should always start at a = 1.
                // Otherwise, the errorSquaredPreviousSteps is tabulated in the 'errorSquared' matrix, and we do a lookup.

                const double sumErrorSquaredPreviousSteps = (a == 1) ? 0.0 : (k == 1) ? INFINITY : sumErrorSquared[(k - 1 - 1) * nMax + (a - 1 - 1)];

                // The total error for a (sub)solution that has the segment [a .. b] as its rightmost segment.

                const double sumErrorSquaredTotal = sumErrorSquaredPreviousSteps + sumErrorSquaredCurrentStep;

                // If the currently considered segmentation is better than the previously considered one (if any),
                // overwrite it.

                if (sumErrorSquaredTotal < sumErrorSquared[(k - 1) * nMax + (b - 1)])
                {
                    sumErrorSquared[(k - 1) * nMax + (b - 1)] = sumErrorSquaredTotal;
                    trace          [(k - 1) * nMax + (b - 1)] = a;
                }
            }
        }
    }

    // If requested, copy the error-squared of the optimal segmentations to the caller.

    if (segSumErrorSquared != NULL)
    {
        for (unsigned k = 1; k <= kMax; ++k)
        {
            segSumErrorSquared[k - 1] = sumErrorSquared[(k - 1) * nMax + (nMax - 1)];
        }
    }

    // If requested, retrieve the optimal segmentations and copy them to the caller.

    if (segBoundaries != NULL)
    {
        for (unsigned k = 1; k <= kMax; ++k)
        {
            for (unsigned nWalk = nMax, kWalk = k; kWalk >= 1; --kWalk)
            {
                segBoundaries[k * (k - 1) / 2 + kWalk - 1] = nWalk;
                nWalk = trace[(kWalk - 1) * nMax + (nWalk - 1)] - 1;
            }
        }
    }

    // Free working arrays

    free(trace);
    free(sumErrorSquared);

    // All done

    return FindOptimalSegmentation_SUCCESS; // report success
}
